
// Various standard headers
#include <iostream>
#include <cstring>
#include <cctype>
#include <cstdlib>
#include <vector>
#include <strstream>
#include <string>
#include <map>
#include <unistd.h>

// Tcl header
#include "tcl.h"
#undef EXTERN

// ROOT headers
#include "TROOT.h"
#include "TFile.h"
#include "TList.h"
#include "TCanvas.h"
#include "TPad.h"
#include "TKey.h"
#include "TAxis.h"
#include "TH1.h"
#include "TH2.h"
#include "TNtuple.h"
#include "TNtupleD.h"

// Histo-Scope declarations
extern "C" {
    void hs_initialize(char *id_string);
    int hs_create_1d_hist(int uid, char *title, char *category,
                          char *x_label, char *y_label,
                          int n_bins, float min, float max);
    int hs_create_2d_hist(int uid, char *title, char *category,
			  char *x_label, char *y_label,
			  char *z_label, int x_bins, int y_bins, float x_min,
			  float x_max, float y_min, float y_max);
    int hs_create_ntuple(int uid, char *title, char *category,
			 int n_variables,char **names);
    void hs_1d_hist_block_fill(int id, float *data, float *err, float *err_m);
    void hs_2d_hist_block_fill(int id, float *data, float *err, float *err_m);
    int hs_fill_ntuple(int id, float *values);
    int hs_save_file(char *name);
}

#define VERSION "1.1"
#define DATE "09/15/2002"
#define AUTHOR "Igor Volobouev, ivolobouev@lbl.gov"

using std::cout;
using std::cerr;
using std::endl;

// List of patterns (global) which will be used by "TraverseObject"
std::vector<std::string> patterns;

// Map from objects into directory names
typedef std::map<TObject *,std::string> DirMap;

// A global flag to allow conversion of doubles into floats
bool allow_reduced_precision = false;


int is_blank_string(const char *pc)
{
    if (pc == NULL)
	return 1;
    for (; *pc; ++pc)
	if (!isspace(*pc))
	    return 0;
    return 1;
}

int GetNextUid(void)
{
    static int i = 0;
    return ++i;
}

char * StringTrim(const char *s)
{
    static char *copy = 0;

    int len = strlen(s);
    delete [] copy;
    copy = new char[len+1];
    strcpy(copy, s);
    if (len == 0)
	return copy;

    // Delete spaces at the end
    char *s_end = copy + len - 1;
    for (; isspace(*s_end) && (s_end > copy); --s_end)
	*s_end = '\0';

    // Skip spaces at the beginning
    char *name = copy;
    for (; isspace(*name) && *name; ++name);

    return name;
}

void TraverseObject(DirMap &hlist, const std::string& dirName, TObject *obj)
{
    const char * classname = obj->ClassName();
    TList *contents = 0;

    if (strcmp("TDirectory", classname) == 0)
    {
	TDirectory *dir = (TDirectory *)obj;
	std::string newDir(dirName);
	if (!is_blank_string(dirName.c_str()))
	    newDir += '/';
	newDir += dir->GetName();
	TList *keylist = dir->GetListOfKeys();
	TListIter *keyiter = new TListIter(keylist);
	TKey *key;
	while ((key = (TKey *)keyiter->Next()))
	{
	    TObject *dirobj = dir->Get(key->GetName());
	    TraverseObject(hlist, newDir, dirobj);
	}
	delete keyiter;
    }
    else if (strcmp("TCanvas", classname) == 0)
    {
        TCanvas *canv = (TCanvas *)obj;
        contents = canv->GetListOfPrimitives();
    }
    else if (strcmp("TPad", classname) == 0)
    {
        TPad *pad = (TPad *)obj;
        contents = pad->GetListOfPrimitives();
    }
    else if (strncmp("TH1", classname, 3) == 0 || 
             strncmp("TH2", classname, 3) == 0 ||
             strcmp("TNtuple", classname) == 0 ||
	     (strcmp("TNtupleD", classname) == 0 && allow_reduced_precision))
    {
	bool matchFound = false;	
	if (patterns.size() > 0)
	{
	    for (size_t i=0; i<patterns.size(); ++i)
	    {
		std::string title(obj->GetName());
		if (Tcl_StringMatch(const_cast<char *>(title.c_str()),
				    const_cast<char *>(patterns[i].c_str())))
		{
		    matchFound = true;
		    break;
		}
	    }
	}
	else
	{
	    matchFound = true;
	}
	if (matchFound)
	{
	    cout << classname << " \"" << obj->GetName() << "\"" << endl;
	    hlist[obj] = dirName;
	}
	else
	{
	    // This object is not going to be useful
	    delete obj;
	}
    }
    if (contents)
    {
        TListIter *iter = new TListIter(contents);
        TObject *item;
        while ((item = iter->Next()))
            TraverseObject(hlist, dirName, item);
	delete iter;
    }
}

DirMap * HistoList(const char *filename)
{
    // Open the input file
    TFile *file = new TFile(filename, "READ");
    if (!file->IsOpen())
    {
        cerr << "ERROR in HistoList: can't open file " 
                  << filename
                  << " for reading"
                  << endl;
        return 0;
    }
    // Traverse the file and find all histograms
    TList *keylist = file->GetListOfKeys();
    DirMap *hlist = new DirMap;
    TListIter *keyiter = new TListIter(keylist);
    TKey *key;
    std::string topdir("");
    while ((key = (TKey *)keyiter->Next()))
    {
        TObject * obj = file->Get(key->GetName());
        TraverseObject(*hlist, topdir, obj);
    }
    delete keyiter;
    return hlist;
}

int ConvertHistogram(TObject *item, char *categ,
		     const std::string& rootDirName,
		     bool remove0sfrom2d)
{
    char *emptystring = "";
    std::ostrstream os;
    
    bool blankPrefix = is_blank_string(categ);
    if (!blankPrefix)
	os << categ;
    if (!is_blank_string(rootDirName.c_str()))
    {
	if (!blankPrefix)
	    os << '/';
	os << rootDirName.c_str();
    }
    os << std::ends;
    char *category = const_cast<char *>(os.str());

    int hs_id = -1;

    // Don't want to use dynamic casts to figure out
    // the class. ROOT already has a funny class called
    // TH1K derived from TH1 which isn't exactly
    // a histogram. Who knows what kind of mutants
    // will be derived from TH1 and TH2 in the future...
    const char * classname = item->ClassName();
    if (strcmp(classname, "TH1C") == 0 ||
	(strcmp(classname, "TH1D") == 0 && allow_reduced_precision) ||
	strcmp(classname, "TH1F") == 0 ||
	strcmp(classname, "TH1S") == 0)
    {
	// This is a 1d histogram
	TH1 *h = dynamic_cast<TH1 *>(item);

	char *x_label;
	if (h->GetXaxis()->GetTitle())
	    x_label = const_cast<char *>(h->GetXaxis()->GetTitle());
	else
	    x_label = emptystring;

	char *y_label;
	if (h->GetYaxis()->GetTitle())
	    y_label = const_cast<char *>(h->GetYaxis()->GetTitle());
	else
	    y_label = emptystring;

	int nbins = h->GetNbinsX();
	hs_id = hs_create_1d_hist(GetNextUid(), StringTrim(h->GetName()), category,
				  x_label, y_label, nbins, 
				  h->GetXaxis()->GetXmin(), h->GetXaxis()->GetXmax());
	if (hs_id > 0)
	{
	    float *data = (float *)malloc(nbins*sizeof(float));
	    if (data == 0)
	    {
		cerr << "Out of memory" << endl;
		exit(EXIT_FAILURE);
	    }
	    for (int ibin = 0; ibin < nbins; ++ibin)
		data[ibin] = h->GetBinContent(ibin+1);
	    hs_1d_hist_block_fill(hs_id, data, 0, 0);
	    free(data);
	}
    }
    else if (strcmp(classname, "TH2C") == 0 ||
	     (strcmp(classname, "TH2D") == 0 && allow_reduced_precision) ||
	     strcmp(classname, "TH2F") == 0 ||
	     strcmp(classname, "TH2S") == 0)
    {
	// This is a 2d histogram
	TH2 *h = dynamic_cast<TH2 *>(item);

	char *x_label;
	if (h->GetXaxis()->GetTitle())
	    x_label = const_cast<char *>(h->GetXaxis()->GetTitle());
	else
	    x_label = emptystring;

	char *y_label;
	if (h->GetYaxis()->GetTitle())
	    y_label = const_cast<char *>(h->GetYaxis()->GetTitle());
	else
	    y_label = emptystring;

	char *z_label;
	if (h->GetZaxis()->GetTitle())
	    z_label = const_cast<char *>(h->GetZaxis()->GetTitle());
	else
	    z_label = emptystring;

	// Get the data
	int x_bins = h->GetNbinsX();
	int y_bins = h->GetNbinsY();
	float *data = (float *)malloc(x_bins*y_bins*sizeof(float));
	if (data == 0)
	{
	    cerr << "Out of memory" << endl;
	    exit(EXIT_FAILURE);
	}
	for (int xbin = 0; xbin < x_bins; ++xbin)
	    for (int ybin = 0; ybin < y_bins; ++ybin)
		data[xbin*y_bins + ybin] = h->GetCellContent(xbin+1, ybin+1);

	// Find out the Histo-Scope histogram limits and the number of bins.
	// Trim the X axis range if necessary.
	int xbin_min = 0;
	int xbin_max = x_bins;
	float xvalue_min = h->GetXaxis()->GetXmin();
	float xvalue_max = h->GetXaxis()->GetXmax();
	if (remove0sfrom2d)
	{
	    if (strncmp(h->GetName(), "histo strip # vs ADC Counts nevt", 32) == 0)
	    {
		bool allZeros = true;
		for (int xbin = 0; xbin < x_bins && allZeros; ++xbin)
		    for (int ybin = 0; ybin < y_bins && allZeros; ++ybin)
			if (data[xbin*y_bins + ybin] != 0.f)
			{
			    xbin_min = xbin;
			    allZeros = false;
			}
		if (allZeros)
		{
		    xbin_min = 0;
		    xbin_max = 2;
		    xvalue_min = 0.f;
		    xvalue_max = 2.f;
		}
		else
		{
		    allZeros = true;
		    for (int xbin = x_bins-1; xbin >= 0 && allZeros; --xbin)
			for (int ybin = 0; ybin < y_bins && allZeros; ++ybin)
			    if (data[xbin*y_bins + ybin] != 0.f)
			    {
				xbin_max = xbin + 1;
				allZeros = false;
			    }
		    if (xbin_min == xbin_max - 1)
		    {
			if (xbin_min)
			    --xbin_min;
			else
			    ++xbin_max;
		    }

		    // Recalculate the histogram range
		    float binwidth = (xvalue_max - xvalue_min)/x_bins;
		    xvalue_max = xvalue_min + xbin_max*binwidth;
		    xvalue_min += xbin_min*binwidth;
		}
	    }
	}

	// Create the Histo-Scope histogram
	hs_id = hs_create_2d_hist(GetNextUid(), StringTrim(h->GetName()), category,
				  x_label, y_label, z_label, xbin_max-xbin_min, y_bins,
				  xvalue_min, xvalue_max,
				  h->GetYaxis()->GetXmin(), h->GetYaxis()->GetXmax());
	if (hs_id > 0)
	    hs_2d_hist_block_fill(hs_id, data + xbin_min*y_bins, 0, 0);

	free(data);
    }
    else if (strcmp(classname, "TNtuple") == 0)
    {
	// This is an Ntuple
	TNtuple *ntuple = (TNtuple *)item;
	int n_vars = ntuple->GetNvar();
	char **names = (char **)calloc(n_vars, sizeof(char *));
	TObjArray *branches = ntuple->GetListOfBranches();
	int i;
	for (i=0; i<n_vars; ++i)
	{
	    TBranch *branch = (TBranch *)branches->UncheckedAt(i);
	    names[i] = strdup(branch->GetName());
	}
	hs_id = hs_create_ntuple(GetNextUid(), StringTrim(ntuple->GetName()), 
				 category, n_vars, names);
	for (i=0; i<n_vars; ++i)
	    free(names[i]);
	free(names);
	if (hs_id > 0)
	{
	    // Fill the ntuple
	    int n_rows = static_cast<int>(ntuple->GetEntries());
	    for (int j=0; j<n_rows; ++j)
	    {
		ntuple->GetEntry(j);
		if (hs_fill_ntuple(hs_id, ntuple->GetArgs()) < 0)
		{
		    cerr << "Out of memory" << endl;
		    exit(EXIT_FAILURE);
		}
	    }
	}
    }
    else if (strcmp(classname, "TNtupleD") == 0 && allow_reduced_precision)
    {
	TNtupleD *ntuple = (TNtupleD *)item;
	int n_vars = ntuple->GetNvar();
	char **names = (char **)calloc(n_vars, sizeof(char *));
	TObjArray *branches = ntuple->GetListOfBranches();
	int i;
	for (i=0; i<n_vars; ++i)
	{
	    TBranch *branch = (TBranch *)branches->UncheckedAt(i);
	    names[i] = strdup(branch->GetName());
	}
	hs_id = hs_create_ntuple(GetNextUid(), StringTrim(ntuple->GetName()), 
				 category, n_vars, names);
	for (i=0; i<n_vars; ++i)
	    free(names[i]);
	free(names);
	if (hs_id > 0)
	{
	    // Fill the ntuple
	    float *fdata = new float[n_vars];
	    int n_rows = static_cast<int>(ntuple->GetEntries());
	    for (int j=0; j<n_rows; ++j)
	    {
		ntuple->GetEntry(j);
		double *ddata = ntuple->GetArgs();
		for (i=0; i<n_vars; ++i)
		    fdata[i] = ddata[i];
		if (hs_fill_ntuple(hs_id, fdata) < 0)
		{
		    cerr << "Out of memory" << endl;
		    exit(EXIT_FAILURE);
		}
	    }
	    delete [] fdata;
	}
    }

    return hs_id;
}

int main(int argc, char *argv[])
{
    char *progname = strrchr(argv[0], '/');
    if (progname)
	++progname;
    else
	progname = argv[0];

    // Arguments which may be specified on the command line
    char *infile = 0;
    char *ofile = 0;
    char *category = 0;
    bool printhelp = false;
    bool printversion = false;
    bool listonly = false;
    bool remove0sfrom2d = false;

    // Variables needed for 'getopt'
    int c;
    bool errflg = false;
    extern char *optarg;
    while((c = getopt(argc, argv, "c:dhi:lo:p:sv")) != EOF)
    {
	switch(c)
	{
	case 'c':
	    category = optarg;
	    break;
	case 'd':
	    allow_reduced_precision = true;
	    break;
	case 'h':
	    printhelp = true;
	    break;
	case 'i':
	    infile = optarg;
	    break;	    
	case 'l':
	    listonly = true;
	    break;	    
	case 'o':
	    ofile = optarg;
	    break;
	case 'p':
	    patterns.push_back(optarg);
	    break;
	case 's':
	    remove0sfrom2d = true;
	    break;
	case 'v':
	    printversion = true;
	    break;
	default:
	    errflg = true;
	    break;
	}
    }

    // Check the arguments
    if (printversion)
    {
	cout << progname << " version " << VERSION 
	     << "  Root to Histo-Scope Converter\n"
	     << "Author: " << AUTHOR << '\n'
	     << "Date: " << DATE
	     << endl;
	if (argc == 2)
	    exit(EXIT_SUCCESS);
    }
    if (printhelp || errflg || infile == 0)
    {
	cout << "\nUsage: "
	     << progname
	     << " -i infile"
	     << " [-c category]"
	     << " [-d]"
	     << " [-h]"
	     << " [-l]"
	     << " [-o ofile]"
	     << " [-s]"
	     << " [-v]\n"
	     << "       [-p pattern1]"
	     << " [-p pattern2]"
	     << " ..."
	     << "\n\n";
	cout << "The following command line switches are supported:\n\n";
	cout << "  -i    infile is the name of the input file\n\n";
	cout << "  -c    category is the name of the Histo-Scope category in which histograms\n";
        cout << "        and histogram directories will be placed. Use this option if you are\n";
        cout << "        going to analyze more than one Histo-Scope file at a time. If this\n";
        cout << "        option is not specified, all histograms and directories will be placed\n";
	cout << "        in the top-level category.\n\n";
	cout << "  -d    allow conversion of doubles into floats\n\n";
	cout << "  -h    print program usage info and exit\n\n";
	cout << "  -l    print the list of histograms in the root file but do not convert\n\n";
        cout << "  -o    ofile is the name of the output file. If this option is not specified,\n";
	cout << "        the output file name will be the same as the input file name but with\n"; 
	cout << "        extension \".hs\".\n\n";
        cout << "  -p    if at least one \"-p\" option is given then the histogram titles will be\n";
	cout << "        required to match the patterns as in the Tcl command \"string match\"\n\n";
        cout << "  -s    be \"smart\" and suppress ADC count regions filled with 0s in 2-d\n";
        cout << "        pulse height distributions.\n\n";
	cout << "  -v    print program verion\n";
        cout << endl;
	if (printhelp)
	    exit(EXIT_SUCCESS);
	else
	    exit(EXIT_FAILURE);
    }

    // Initialize root
    TROOT root("root2hs", "Root to Histo-Scope Converter");

    // Get the list of histograms
    DirMap *hlist = HistoList(infile);
    if (hlist == 0)
    {
	// HistoList has printed an error message already
	exit(EXIT_FAILURE);
    }
    else if (hlist->size() == 0)
    {
	cout << "No histograms found in file \"" 
		  << infile << "\"" << endl;
	exit(EXIT_FAILURE);
    }

    // Check if only histogram list is requested
    if (listonly)
    {
	// We are done here
	exit(EXIT_SUCCESS);
    }

    // Initialize Histo-Scope
    hs_initialize("Root to Histo-Scope Converter");

    // Cycle over all ROOT histograms and create Histo-Scope analogs
    int num_converted = 0;
    for (DirMap::const_iterator iter = hlist->begin();
	 iter != hlist->end(); ++iter)
    {
	int hs_id = ConvertHistogram(iter->first, category,
				     iter->second, remove0sfrom2d);
	if (hs_id <= 0)
	    cout << "ERROR: failed to convert " << iter->first->ClassName()
		 << " \"" << iter->first->GetName() << "\"" << endl;
	else
	    ++num_converted;
	
	// Important -- delete the root item in memory. Otherwise
        // the memory usage will be essentially doubled.
	delete iter->first;
    }
    if (num_converted == 0)
    {
	cout << "No histograms converted from file \"" 
		  << infile << "\"" << endl;
	exit(EXIT_FAILURE);
    }

    // Save the Histo-Scope file
    if (ofile == 0)
    {
	// Generate an output file name
	ofile = (char *)malloc(strlen(infile) + strlen(".hs") + 1);
	char *tmp = strrchr(infile, '/');
	if (tmp)
	    strcpy(ofile, tmp+1);
	else
	    strcpy(ofile, infile);
	tmp = strrchr(ofile, '.');
	if (tmp)
	    *tmp = '\0';
	strcat(ofile, ".hs");
    }
    
    int num_items = hs_save_file(ofile);
    cout << "Wrote " << num_items << " histograms to file "
	 << ofile << endl;
    
    if (num_items > 0)
	exit(EXIT_SUCCESS);
    else
	exit(EXIT_FAILURE);
}
